import os
import traceback
import socket

import win32security
from win32api import CloseHandle

from utils.meta import (
    WMIMetaHandler,
    win_cli_output_decoder,
    get_remote_apps,
)
from models.meta import (
    MetaUserModel,
    MetaLogoutModel,
)

STATE_ACTIVE = ('Активно', 'Active')
SESSION_STATES = "IA"


class MetaService:
    def get_sessions(self) -> dict:
        """function returning active rds sessions.

        Returns:
            result (dict): json response
        """
        try:
            result = []
            output = os.popen("query user").readlines()
            # Each line has the following format:
            # user4        rdp-tcp#15     5  Active   6:42  17.08.2022 10:16
            #   1              2          3     4       5        6       7
            # 1 - Username
            # 2 - Name of session
            # 3 - Session id
            # 4 - Current session state
            # 5 - Session idle time. If the session is in use, then the time will be "."
            # 6 - Session creation date
            # 7 - Session creation time

            remote_app_sessions_id = WMIMetaHandler().get_remote_app_sessions_id()
            # Можно отделить локальные сессии от rdp при помощи списка всех сессий с процессом "rdpclip.exe"

            # fqdn_server = socket.getfqdn()  # For future
            host_server = socket.gethostbyname(socket.gethostname())

            # Start enumeration without first element, because this is the title
            for line in output[1:]:
                elements = win_cli_output_decoder(line)
                elements = elements.split(" ")

                # Original string contains spaces for formatting, which after split() give unnecessary
                # empty strings, which we exclude from the list
                elements[:] = filter(lambda x: x != "", elements)

                # Current username leads with extra mark '>'. Exclude it
                username = elements[0].replace(">", "")

                # Some sessions may not have a session name. This is why we check
                session_id = int(elements[2] if elements[2].isdigit() else elements[1])
                state = elements[3] if elements[2].isdigit() else elements[4]

                result.append(
                    {
                        "username": username,
                        "session_id": session_id,
                        "host_server": host_server,  # for logoff in future
                        "state": SESSION_STATES[state in STATE_ACTIVE],
                        "is_desktop": session_id not in remote_app_sessions_id,
                    }
                )
            result = {"sessions": result, "error": None}
        except Exception as ex:
            result = {
                "sessions": None,
                "error": str(ex),
                "traceback": traceback.format_exc(),
            }

        return result

    def login(self, params: MetaUserModel) -> dict:
        """authorization function in rds

        Args:
            params (RdsInputModel): auth params

        Returns:
            result (dict): json response
        """

        try:
            # LogonUser() return a handle to a token that represents the specified user.
            # Returned handle is a primary token that you can use in calls to the
            # CreateProcessAsUser() function.
            # If you specify the LOGON32_LOGON_NETWORK flag, LogonUser() returns
            # an impersonation token that you cannot use in CreateProcessAsUser()
            # unless you call DuplicateTokenEx() to convert it to a primary token.
            # LOGON32_PROVIDER_WINNT40: Use the NTLM logon provider.
            token = win32security.LogonUser(
                params.login,
                params.domain,
                params.password,
                win32security.LOGON32_LOGON_NETWORK,
                win32security.LOGON32_PROVIDER_WINNT40,
            )
            authenticated = bool(token)
            CloseHandle(token)

            result = dict(auth_done=authenticated, error=None)
        except Exception as ex:
            result = dict(
                auth_done=False,
                error=str(ex),
                traceback=traceback.format_exc(),
            )

        return result

    def connect(self, params: MetaUserModel) -> dict:
        """method that return response of trying to connect to rds

        Args:
            params (RdsUserModel): connect params

        Returns:
            result (dict): json response
        """
        try:
            # TODO Пока не понятно как правильно делать проверку тестового подключения по RDP.
            # Оставим текущую логику закомментированной.

            # RDS_HOSTNAME = '192.168.3.13'
            # login = params.login
            # pwd = params.password

            # os.system(
            #     'cmd/c "cmdkey /generic:"{hostname}" /user:"{login}" \
            #         /pass:"{pwd}""'.format(hostname=RDS_HOSTNAME, login=login, pwd=pwd)
            #     )
            # os.system('cmd/c "mstsc /v:{hostname}"'.format(hostname=RDS_HOSTNAME))
            # os.system('cmd/c "cmdkey /delete:{hostname}"'.format(hostname=RDS_HOSTNAME))

            # Временное решение делать проверку посредством попытки авторизации на AD.
            rds_login_response = self.login(params)
            result = {"is_connected": True, "error": None}
            if not rds_login_response["auth_done"]:
                result = {
                    "is_connected": False,
                    "error": str(rds_login_response["error"]),
                }
        except Exception as ex:
            result = {
                "is_connected": False,
                "error": str(ex),
                "traceback": traceback.format_exc(),
            }

        return result

    def logout_user(self, params: MetaLogoutModel) -> dict:
        """method disconnect user from RDS

        Args:
            params (RdsInputModel): logout params

        Returns:
            result (dict): json response
        """
        try:
            session_id = params.session_id
            # In future use host server
            # Example: logoff /VM <session id> /server:<host server>
            os.system(
                'cmd/c "echo y|LOGOFF /VM {sessin_id}"'.format(sessin_id=session_id)
            )

            result = dict(disconnect_done=True, error=None)
        except Exception as ex:
            result = dict(
                disconnect_done=False,
                error=str(ex),
                traceback=traceback.format_exc(),
            )

        return result

    def get_remote_apps(self) -> dict:
        """Getting published remote apps from RDS by collection name.

        Args:
            params (RdsRemoteAppModel): contains collection name

        Returns:
            result (dict): json response
        """
        try:
            result = {"apps": get_remote_apps(), "error": None}

        except Exception as ex:
            result = dict(
                apps=None,
                error=str(ex),
                traceback=traceback.format_exc(),
            )

        return result
